from ctypes import *
import numpy as np
from . import corelib

# Define the call signatures
_compute_mse = corelib.core.ComputeMSE
_compute_mse.argtypes = [POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int ]
_compute_mse.restype = c_float

_compute_rel_mse = corelib.core.ComputeRelMSE
_compute_rel_mse.argtypes = [POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int ]
_compute_rel_mse.restype = c_float

_compute_rel_mse_outlier_reject = corelib.core.ComputeRelMSEOutlierReject
_compute_rel_mse_outlier_reject.argtypes = [
    POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int, c_float ]
_compute_rel_mse_outlier_reject.restype = c_float

def mse(img, ref):
    return corelib.invoke_on_pair(_compute_mse, img, ref)

def relative_mse(img, ref):
    return corelib.invoke_on_pair(_compute_rel_mse, img, ref)

def relative_mse_outlier_rejection(img, ref, percentage=0.1):
    return corelib.invoke_on_pair(_compute_rel_mse_outlier_reject, img, ref, percentage)


from . import corelib
from ctypes import *
import numpy as np

_reinhard = corelib.core.TonemapReinhard
_reinhard.argtypes = [POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int, c_float ]
_reinhard.restype = c_float

_aces = corelib.core.TonemapACES
_aces.argtypes = [POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int ]
_aces.restype = c_float

def reinhard(img, max_luminance):
    return corelib.invoke_with_output(_reinhard, img, max_luminance)

def aces(img):
    return corelib.invoke_with_output(_aces, img)


from . import corelib
from ctypes import *
import numpy as np

_exposure = corelib.core.AdjustExposure
_exposure.argtypes = (POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int, c_float)
_exposure.restype = None

_lin_to_srgb = corelib.core.LinearToSrgb
_lin_to_srgb.argtypes = (POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int)
_lin_to_srgb.restype = None

_srgb_to_lin = corelib.core.SrgbToLinear
_srgb_to_lin.argtypes = (POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int)
_srgb_to_lin.restype = None

_to_byte_img = corelib.core.ToByteImage
_to_byte_img.argtypes = (POINTER(c_float), c_int, POINTER(c_uint8), c_int, c_int, c_int, c_int)
_to_byte_img.restype = None

_zoom = corelib.core.ZoomWithNearestInterp
_zoom.argtypes = (POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int, c_int)
_zoom.restype = None

_average_to_mono = corelib.core.RgbToMonoAverage
_average_to_mono.argtypes = (POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int)
_average_to_mono.restype = None

_lum_to_mono = corelib.core.RgbToMonoLuminance
_lum_to_mono.argtypes = (POINTER(c_float), c_int, POINTER(c_float), c_int, c_int, c_int, c_int)
_lum_to_mono.restype = None

def exposure_inplace(img, exposure=0):
    img = np.array(img, dtype=np.float32, copy=False)
    h = img.shape[0]
    w = img.shape[1]
    _exposure(img.ctypes.data_as(POINTER(c_float)), img.ctypes.data_as(POINTER(c_float)), w, h, exposure)
    return img

def exposure(img, exposure=0):
    return corelib.invoke_with_output(_exposure, img, exposure)

def lin_to_srgb(img):
    return corelib.invoke_with_output(_lin_to_srgb, img)

def srgb_to_lin(img):
    return corelib.invoke_with_output(_srgb_to_lin, img)

def to_byte_image(img):
    return corelib.invoke_with_byte_output(_to_byte_img, img)

def zoom(img, scale: int):
    img = np.array(img, dtype=np.float32, copy=False)
    h = img.shape[0]
    w = img.shape[1]
    if len(img.shape) < 3:
        buf = np.zeros((h * scale, w * scale), dtype=np.float32)
    else:
        buf = np.zeros((h * scale, w * scale, img.shape[2]), dtype=np.float32)
    corelib.invoke_on_pair(_zoom, img, buf, scale)
    return buf

def average_color_channels(img):
    img = np.array(img, dtype=np.float32, copy=False)
    h = img.shape[0]
    w = img.shape[1]
    buf = np.zeros((h, w), dtype=np.float32)
    corelib.invoke_on_pair(_average_to_mono, img, buf)
    return buf

def luminance(img):
    img = np.array(img, dtype=np.float32, copy=False)
    h = img.shape[0]
    w = img.shape[1]
    buf = np.zeros((h, w), dtype=np.float32)
    corelib.invoke_on_pair(_lum_to_mono, img, buf)
    return buf


# Blatantly stolen from the excellent gist by Tomáš Iser (https://cgg.mff.cuni.cz/~tomas/):
# https://gist.github.com/tomasiser/5e3bacd72df30f7efc3037cb95a039d3
# Adapted slightly to better fit the other parts of SimpleImageIO, especially the image representation

from . import corelib
import socket
import struct

class TevIpc:
    def __init__(self, hostname = "localhost", port = 14158):
        self._hostname = hostname
        self._port = port
        self._socket = None

    def __enter__(self):
        if self._socket is not None:
            raise Exception("Communication already started")
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM) # SOCK_STREAM means a TCP socket
        self._socket.__enter__()
        self._socket.connect((self._hostname, self._port))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._socket is None:
            raise Exception("Communication was not started")
        self._socket.__exit__(exc_type, exc_val, exc_tb)

    def create_image(self, name: str, width: int, height: int, channel_names, grab_focus = True):
        if self._socket is None:
            raise Exception("Communication was not started")

        data_bytes = bytearray()
        data_bytes.extend(struct.pack("<I", 0)) # reserved for length
        data_bytes.extend(struct.pack("<b", 4)) # create image
        data_bytes.extend(struct.pack("<b", grab_focus)) # grab focus
        data_bytes.extend(bytes(name, "ascii")) # image name
        data_bytes.extend(struct.pack("<b", 0)) # string terminator
        data_bytes.extend(struct.pack("<i", width)) # width
        data_bytes.extend(struct.pack("<i", height)) # height
        data_bytes.extend(struct.pack("<i", len(channel_names))) # number of channels
        for cname in channel_names:
            data_bytes.extend(bytes(cname, "ascii")) # channel name
            data_bytes.extend(struct.pack("<b", 0)) # string terminator
        data_bytes[0:4] = struct.pack("<I", len(data_bytes))

        self._socket.sendall(data_bytes)

    def display_image(self, name: str, image, grab_focus = True):
        data, (stride, width, height, num_channels) = corelib.get_numpy_data(image)
        if num_channels == 1:
            channel_names = ["Y"]
        elif num_channels == 3:
            channel_names = ["R", "G", "B"]
        elif num_channels == 4:
            channel_names = ["R", "G", "B", "A"]
        self.close_image(name)
        self.create_image(name, width, height, channel_names, grab_focus)
        self.update_image(name, image)

    def display_layered_image(self, name: str, layers: dict, grab_focus = True):
        channel_names = []
        for layer_name, image in layers.items():
            data, (stride, width, height, num_channels) = corelib.get_numpy_data(image)
            if num_channels == 1:
                channel_names.extend([f"{layer_name}.Y"])
            elif num_channels == 3:
                channel_names.extend([f"{layer_name}.R", f"{layer_name}.G", f"{layer_name}.B"])
            elif num_channels == 4:
                channel_names.extend([f"{layer_name}.R", f"{layer_name}.G", f"{layer_name}.B", f"{layer_name}.A"])
        self.close_image(name)
        self.create_image(name, width, height, channel_names, grab_focus)
        self.update_layered_image(name, layers)

    def update_image(self, name: str, image, grab_focus = False):
        data, (stride, width, height, num_channels) = corelib.get_numpy_data(image)
        if num_channels == 1:
            channel_names = ["Y"]
        elif num_channels == 3:
            channel_names = ["R", "G", "B"]
        elif num_channels == 4:
            channel_names = ["R", "G", "B", "A"]
        idx = 0
        for c in channel_names:
            channel_data = data[:,:,idx].tobytes()
            self._update_image(name, c, width, height, channel_data, grab_focus)
            idx += 1

    def update_layered_image(self, name: str, layers, grab_focus = False):
        for layer_name, image in layers.items():
            data, (stride, width, height, num_channels) = corelib.get_numpy_data(image)
            if num_channels == 1:
                channel_names = ["Y"]
            elif num_channels == 3:
                channel_names = ["R", "G", "B"]
            elif num_channels == 4:
                channel_names = ["R", "G", "B", "A"]
            idx = 0
            for c in channel_names:
                channel_data = data[:,:,idx].tobytes()
                self._update_image(name, f"{layer_name}.{c}", width, height, channel_data, grab_focus)
                idx += 1

    def _update_image(self, name: str, channel_name: str, width, height, byte_data, grab_focus = False):
        if self._socket is None:
            raise Exception("Communication was not started")

        data_bytes = bytearray()
        data_bytes.extend(struct.pack("<I", 0)) # reserved for length
        data_bytes.extend(struct.pack("<b", 3)) # update image
        data_bytes.extend(struct.pack("<b", grab_focus)) # grab focus
        data_bytes.extend(bytes(name, "ascii")) # image name
        data_bytes.extend(struct.pack("<b", 0)) # string terminator
        data_bytes.extend(bytes(channel_name, "ascii")) # channel name
        data_bytes.extend(struct.pack("<b", 0)) # string terminator
        data_bytes.extend(struct.pack("<i", 0)) # x
        data_bytes.extend(struct.pack("<i", 0)) # y
        data_bytes.extend(struct.pack("<i", width)) # width
        data_bytes.extend(struct.pack("<i", height)) # height
        data_bytes.extend(byte_data) # data
        data_bytes[0:4] = struct.pack("<I", len(data_bytes))

        self._socket.sendall(data_bytes)

    def close_image(self, name: str):
        if self._socket is None:
            raise Exception("Communication was not started")

        data_bytes = bytearray()
        data_bytes.extend(struct.pack("<I", 0)) # reserved for length
        data_bytes.extend(struct.pack("<b", 2)) # close image
        data_bytes.extend(bytes(name, "ascii")) # image name
        data_bytes.extend(struct.pack("<b", 0)) # string terminator
        data_bytes[0:4] = struct.pack("<I", len(data_bytes))

        self._socket.sendall(data_bytes)


import pathlib
import platform
from ctypes import *
import numpy as np

if platform.system() == "Windows":
    _dll_name = "SimpleImageIOCore.dll"
elif platform.system() == "Darwin":
    _dll_name = "libSimpleImageIOCore.dylib"
else: # Use Linux-style and hope for the best :)
    _dll_name = "libSimpleImageIOCore.so"

_libpath = str(pathlib.Path(__file__).with_name(_dll_name))
core = cdll.LoadLibrary(_libpath)

def get_numpy_data(img):
    # Make sure the data type is float32, only copy data if a cast is required
    # This also automatically generates a numpy array in case img was just a plain list of lists
    img = np.array(img, dtype=np.float32, copy=False)

    if len(img.shape) == 3 and (img.strides[2] != 4 or img.strides[1] != 4 * 3):
        # we could have been given a masked array with gaps between values
        # our C-API does not support this, so we need to copy
        img = np.array(img, dtype=np.float32, copy=True)

    height = img.shape[0]
    width = img.shape[1]
    if len(img.shape) == 2:
        num_channels = 1
    else:
        num_channels = img.shape[2]

    assert len(img.shape) == 2 or img.strides[2] == 4
    assert img.strides[1] == 4 * num_channels

    # If we are given a slice that is, e.g., a tile of the image, the stride between rows can differ
    # The C-API expects the stride in multiples of sizeof(float32), which is 4
    stride = int(img.strides[0] / 4)

    # Now the data is in a format that our C-API will understand
    return img, (stride, width, height, num_channels)

def invoke(func, img, *args):
    """
    Calls a C-API function that gets the paramters:
    (float* image, int stride, int width, int height, int numChannels, *args)
    Returns the result, or None
    """
    img, dims = get_numpy_data(img)
    return func(img.ctypes.data_as(POINTER(c_float)), *dims, *args)

def invoke_with_output(func, img, *args):
    """
    Calls a C-API function that gets the paramters:
    (float* image, int strideIn, float* output, int strideOut, int width, int height, int numChannels, *args)
    Returns the numpy array that contains the data written to 'output'
    """
    img, dims = get_numpy_data(img)
    if dims[3] == 1:
        buffer = np.zeros((dims[2], dims[1]), dtype=np.float32)
    else:
        buffer = np.zeros((dims[2], dims[1], dims[3]), dtype=np.float32)
    func(img.ctypes.data_as(POINTER(c_float)), dims[0],
        buffer.ctypes.data_as(POINTER(c_float)), int(buffer.strides[0] / 4),
        dims[1], dims[2], dims[3], *args)
    return buffer

def invoke_with_byte_output(func, img, *args):
    """
    Calls a C-API function that gets the paramters:
    (float* image, int strideIn, float* output, int strideOut, int width, int height, int numChannels, *args)
    Returns the numpy array that contains the data written to 'output'
    """
    img, dims = get_numpy_data(img)
    if dims[3] == 1:
        buffer = np.zeros((dims[2], dims[1]), dtype=np.uint8)
    else:
        buffer = np.zeros((dims[2], dims[1], dims[3]), dtype=np.uint8)
    func(img.ctypes.data_as(POINTER(c_float)), dims[0],
        buffer.ctypes.data_as(POINTER(c_uint8)), buffer.strides[0],
        dims[1], dims[2], dims[3], *args)
    return buffer

def invoke_on_pair(func, first, second, *args):
    """
    Calls a C-API function that gets the paramters:
    (float* image, int strideIn, float* output, int strideOut, int width, int height, int numChannels, *args)
    Returns the result, or None
    """
    img_a, dims_a = get_numpy_data(first)
    img_b, dims_b = get_numpy_data(second)

    return func(img_a.ctypes.data_as(POINTER(c_float)), dims_a[0],
        img_b.ctypes.data_as(POINTER(c_float)), dims_b[0],
        dims_a[1], dims_a[2], dims_a[3], *args)


# The CMake interop code is based paritally on hoefling's excellent answer
# to this StackOverflow question:
# https://stackoverflow.com/questions/42585210/extending-setuptools-extension-to-use-cmake-in-setup-py

import os
import pathlib
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig

class CMakeExtension(Extension):
    def __init__(self, name):
        # don't invoke the original build_ext for this special extension
        super().__init__(name, sources=[])

class build_ext(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)
        super().run()

    def build_cmake(self, ext):
        cwd = pathlib.Path().absolute()

        # these dirs will be created in build_py, so if you don't have
        # any python sources to bundle, the dirs will be missing
        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)
        extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
        extdir.mkdir(parents=True, exist_ok=True)

        # Configure CMake arguments
        config = 'Release'
        cmake_args = [
            '-DPYLIB=' + str(extdir.parent.absolute()), # destination for the shared library
            '-DCMAKE_BUILD_TYPE=' + config
        ]
        build_args = [ '--config', config ]

        # Run CMake and build (automatically copies the .dll / .so / .dylib file)
        os.chdir(str(build_temp))
        self.spawn(['cmake', str(cwd)] + cmake_args)
        if not self.dry_run:
            self.spawn(['cmake', '--build', '.'] + build_args)
        os.chdir(str(cwd))

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='simpleimageio',
    version='0.12.0',
    author='Pascal Grittmann',
    url='https://github.com/pgrit/SimpleImageIO',

    description='Python wrapper around TinyEXR and stb_image',
    long_description=long_description,
    long_description_content_type="text/markdown",

    packages=['simpleimageio'],
    package_dir={'simpleimageio': 'PyWrapper/simpleimageio'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy'
    ],
    ext_modules=[CMakeExtension('simpleimageio/SimpleImageIOCore')],
    cmdclass={
        'build_ext': build_ext,
    }
)



from ctypes import *
import numpy as np
from . import corelib
import base64
import collections

_write_image = corelib.core.WriteImage
_write_image.argtypes = [POINTER(c_float), c_int, c_int, c_int, c_int, c_char_p, c_int]
_write_image.restype = None

_write_layered_exr = corelib.core.WriteLayeredExr
_write_layered_exr.argtypes = [
    POINTER(POINTER(c_float)), POINTER(c_int), c_int, c_int, POINTER(c_int), c_int, POINTER(c_char_p), c_char_p]
_write_layered_exr.restype = None

_cache_image = corelib.core.CacheImage
_cache_image.argtypes = [POINTER(c_int), POINTER(c_int), POINTER(c_int), c_char_p]
_cache_image.restype = c_int

_copy_cached_img = corelib.core.CopyCachedImage
_copy_cached_img.argtypes = [c_int, POINTER(c_float)]
_copy_cached_img.restype = None

_write_to_mem = corelib.core.WriteToMemory
_write_to_mem.argtypes = [POINTER(c_float), c_int, c_int, c_int, c_int, c_char_p, c_int, POINTER(c_int)]
_write_to_mem.restype = POINTER(c_ubyte)

_free_mem = corelib.core.FreeMemory
_free_mem.argtypes = [POINTER(c_ubyte)]
_free_mem.restype = None

_get_layer_count = corelib.core.GetExrLayerCount
_get_layer_count.argtypes = [c_int]
_get_layer_count.restype = c_int

_get_layer_chan_count = corelib.core.GetExrLayerChannelCount
_get_layer_chan_count.argtypes = [c_int, c_char_p]
_get_layer_chan_count.restype = c_int

_get_layer_name_len = corelib.core.GetExrLayerNameLen
_get_layer_name_len.argtypes = [c_int, c_int]
_get_layer_name_len.restype = c_int

_get_layer_name = corelib.core.GetExrLayerName
_get_layer_name.argtypes = [c_int, c_int, POINTER(c_char)]
_get_layer_name.restype = None

_copy_layer = corelib.core.CopyCachedLayer
_copy_layer.argtypes = [c_int, c_char_p, POINTER(c_float)]
_copy_layer.restype = None

_delete_image = corelib.core.DeleteCachedImage
_delete_image.argtypes = [c_int]
_delete_image.restype = None

def read(filename: str):
    w = c_int()
    h = c_int()
    c = c_int()
    idx = _cache_image(byref(w), byref(h), byref(c), filename.encode('utf-8'))
    chans = c.value

    if chans == 1:
        buffer = np.zeros((h.value,w.value), dtype=np.float32)
    else:
        buffer = np.zeros((h.value,w.value,chans), dtype=np.float32)

    _copy_cached_img(idx, buffer.ctypes.data_as(POINTER(c_float)))

    # We enforce rgb results and always drop any extra channels (alpha)
    if chans > 3:
        return buffer[:,:,:3]

    return buffer

def read_layered_exr(filename: str):
    w = c_int()
    h = c_int()
    c = c_int()
    idx = _cache_image(byref(w), byref(h), byref(c), filename.encode('utf-8'))
    num_layer = _get_layer_count(idx)

    layers = {}

    for i in range(num_layer):
        num_char = _get_layer_name_len(idx, i)
        str_buf = create_string_buffer(num_char)
        _get_layer_name(idx, i, str_buf)
        name = str_buf.value

        num_chans = _get_layer_chan_count(idx, name)
        if num_chans == 1:
            buffer = np.zeros((h.value,w.value), dtype=np.float32)
        else:
            buffer = np.zeros((h.value,w.value,num_chans), dtype=np.float32)

        _copy_layer(idx, name, buffer.ctypes.data_as(POINTER(c_float)))

        layers[name.decode('utf-8')] = buffer

    _delete_image(idx)

    return layers

def write(filename: str, data, jpeg_quality = 80):
    corelib.invoke(_write_image, data, filename.encode('utf-8'), jpeg_quality)

def write_layered_exr(filename: str, layers: dict):
    names = sorted(layers.keys())

    # Gather the data in the desired layout for the C-API
    num_layers = len(layers)
    images = []
    strides = []
    strides = []
    width = -1
    height = -1
    num_channels = []
    cstr_names = []
    for name in names:
        img = layers[name]
        buffer, (stride, w, h, c) = corelib.get_numpy_data(img)
        images.append(buffer.ctypes.data_as(POINTER(c_float)))
        strides.append(stride)
        assert width == -1 or w == width, "all layers must have the same resolution"
        assert height == -1 or h == height, "all layers must have the same resolution"
        width = w
        height = h
        num_channels.append(c)
        cstr_names.append(name.encode('utf-8'))

    _write_layered_exr((POINTER(c_float) * num_layers)(*images),
        (c_int * num_layers)(*strides), width, height,
        (c_int * num_layers)(*num_channels), num_layers, (c_char_p * num_layers)(*cstr_names),
        filename.encode('utf-8'))

def base64_png(img):
    numbytes = c_int()
    mem = corelib.invoke(_write_to_mem, img, ".png".encode('utf-8'), 0, byref(numbytes))
    b64 = base64.b64encode(bytearray(mem[:numbytes.value]))
    _free_mem(mem)
    return b64

def base64_jpg(img, quality = 80):
    numbytes = c_int()
    mem = corelib.invoke(_write_to_mem, img, ".jpg".encode('utf-8'), quality, byref(numbytes))
    b64 = base64.b64encode(bytearray(mem[:numbytes.value]))
    _free_mem(mem)
    return b64